@@ -1,7 +1,7 @@
+"""PyAsusWRT Library implementation"""
 
 from __future__ import annotations
 
-from aiohttp import ClientSession
 import asyncio
 import base64
 from collections import namedtuple
@@ -9,6 +9,9 @@ from datetime import datetime
 import json
 import logging
 import math
+from typing import Any, Optional
+
+import aiohttp
 
 ASUSWRT_USR_AGENT = "asusrouter-Android-DUTUtil-1.0.0.245"
 ASUSWRT_ERROR_KEY = "error_status"
@@ -32,16 +35,21 @@ DEFAULT_TIMEOUT = 5
 DEFAULT_HTTP_PORT = 80
 DEFAULT_HTTPS_PORT = 8443
 
+PROP_MAC_ADDR = "label_mac"
+
 NVRAM_INFO = [
     "acs_dfs",
     "model",
     "productid",
-    "label_mac",
-    "firmver",
-    "innerver",
+    PROP_MAC_ADDR,
     "buildinfo",
+    "firmver",
+    "firmver_org",
     "buildno",
     "buildno_org",
+    "extendno",
+    "extendno_org",
+    "innerver",
     "apps_sq",
     "lan_hwaddr",
     "lan_ipaddr",
@@ -53,7 +61,7 @@ NVRAM_INFO = [
     "https_lanport",
     "cfg_device_list",
     "wl0_country_code",
-    "wl1_country_code"
+    "wl1_country_code",
     "time_zone",
     "time_zone_dst",
     "time_zone_x",
@@ -62,7 +70,7 @@ NVRAM_INFO = [
     "ntp_server0",
 ]
 
-Device = namedtuple("Device", ["mac", "ip", "name"])
+Device = namedtuple("Device", ["mac", "ip", "name", "node", "is_mesh_node", "is_wl"])
 
 _LOGGER = logging.getLogger(__name__)
 
@@ -72,28 +80,89 @@ def _nvram_cmd(info_type):
     return f"{CMD_NVRAM}({info_type})"
 
 
-class AsusWrtConnectionError(Exception):
-    """Error communicating with the router."""
-    pass
+def _get_json_result(result: str, json_key: str | None = None):
+    """Return the json result from a text result."""
+    try:
+        json_res = json.loads(result)
+    except json.JSONDecodeError as exc:
+        raise AsusWrtValueError(str(exc)) from exc
+
+    if not json_key:
+        return json_res
+
+    if (json_val := json_res.get(json_key)) is None:
+        raise AsusWrtValueError("No value available")
+    return json_val
 
 
-class AsusWrtLoginError(Exception):
+class AsusWrtError(Exception):
+    """Base class for all errors raised by this library."""
+
+    def __init__(
+        self, *args: Any, message: Optional[str] = None, **_kwargs: Any
+    ) -> None:
+        """Initialize base AsusWrtError."""
+        super().__init__(*args, message)
+
+
+class AsusWrtCommunicationError(AsusWrtError, aiohttp.ClientError):
+    """Error occurred while communicating with the AsusWrt device ."""
+
+
+class AsusWrtResponseError(AsusWrtCommunicationError):
+    """HTTP error code returned by the AsusWrt device."""
+
+    def __init__(
+        self,
+        *args: Any,
+        status: int,
+        headers: Optional[aiohttp.typedefs.LooseHeaders] = None,
+        message: Optional[str] = None,
+        **kwargs: Any,
+    ) -> None:
+        """Initialize."""
+        if not message:
+            message = f"Did not receive HTTP 200 but {status}"
+        super().__init__(*args, message=message, **kwargs)
+        self.status = status
+        self.headers = headers
+
+
+class AsusWrtClientResponseError(aiohttp.ClientResponseError, AsusWrtResponseError):
+    """HTTP response error with more details from aiohttp."""
+
+
+class AsusWrtConnectionError(AsusWrtCommunicationError, aiohttp.ClientConnectionError):
+    """Error connecting with the router."""
+
+
+class AsusWrtConnectionTimeoutError(
+    AsusWrtCommunicationError, aiohttp.ServerTimeoutError, asyncio.TimeoutError
+):
+    """Timeout while communicating with the device."""
+
+
+class AsusWrtLoginError(AsusWrtError):
     """Login error / invalid credential."""
-    pass
+
+
+class AsusWrtValueError(AsusWrtError, ValueError):
+    """Error invalid value received."""
 
 
 class AsusWrtHttp:
+    """Class for AsusWrt router HTTP/HTTPS connection."""
 
     def __init__(
-            self,
-            hostname: str,
-            username: str,
-            password: str,
-            *,
-            use_https: bool = False,
-            port: int | None = None,
-            timeout: int = DEFAULT_TIMEOUT,
-            session: ClientSession | None = None
+        self,
+        hostname: str,
+        username: str,
+        password: str,
+        *,
+        use_https: bool = False,
+        port: int | None = None,
+        timeout: int = DEFAULT_TIMEOUT,
+        session: aiohttp.ClientSession | None = None,
     ):
         """
         Create the router object
@@ -122,9 +191,10 @@ class AsusWrtHttp:
             self._session = session
             self._managed_session = False
         else:
-            self._session = ClientSession()
+            self._session = None
             self._managed_session = True
 
+        self._mac = None
         self._latest_transfer_data = None
         self._latest_transfer_rate = {"rx_rate": 0.0, "tx_rate": 0.0}
         self._latest_transfer_check = None
@@ -133,76 +203,66 @@ class AsusWrtHttp:
         """Return the url to a specific path."""
         return f"{self._protocol}://{self._hostname}:{self._port}/{path}"
 
-    @property
-    def is_connected(self) -> bool:
-        """Return if connection is active."""
-        return self._auth_headers is not None
-
-    async def async_disconnect(self):
-        """Close the managed session on exit."""
-        if self._managed_session:
-            await self._session.close()
-        self._auth_headers = None
-
-    async def async_connect(self):
-        """Authenticate with the router."""
-        if self.is_connected:
-            return
-
-        auth = f"{self._username}:{self._password}".encode("ascii")
-        login_token = base64.b64encode(auth).decode("ascii")
-        payload = f"login_authorization={login_token}"
-        headers = {"user-agent": ASUSWRT_USR_AGENT}
-
+    async def __http_post(self, url, headers, payload, *, get_json=False):
+        """Perform aiohttp POST request."""
         try:
             async with self._session.post(
-                url=self.__url(ASUSWRT_LOGIN_PATH),
+                url=url,
                 headers=headers,
                 data=payload,
                 timeout=self._timeout,
                 raise_for_status=True,
                 ssl=False,
             ) as resp:
-                result = await resp.json()
-        except (asyncio.TimeoutError, ConnectionRefusedError, OSError) as exc:
-            raise AsusWrtConnectionError(exc) from exc
+                if get_json:
+                    result = await resp.json()
+                else:
+                    result = await resp.text()
 
-        if ASUSWRT_TOKEN_KEY not in result:
-            raise AsusWrtLoginError()
+        except (asyncio.TimeoutError, aiohttp.ServerTimeoutError) as err:
+            raise AsusWrtConnectionTimeoutError(str(err)) from err
+        except aiohttp.ClientConnectorError as err:
+            raise AsusWrtConnectionError(str(err)) from err
+        except aiohttp.ClientConnectionError as err:
+            self._auth_headers = None
+            raise AsusWrtConnectionError(str(err)) from err
+        except aiohttp.ClientResponseError as err:
+            raise AsusWrtClientResponseError(
+                request_info=err.request_info,
+                history=err.history,
+                status=err.status,
+                message=err.message,
+                headers=err.headers,
+            ) from err
+        except aiohttp.ClientError as err:
+            self._auth_headers = None
+            raise AsusWrtError(str(err)) from err
 
-        token = result[ASUSWRT_TOKEN_KEY]
-        self._auth_headers = {
-            "user-agent": ASUSWRT_USR_AGENT,
-            "cookie": f"{ASUSWRT_TOKEN_KEY}={token}",
-        }
+        return result
 
-    async def __post(self, command, path=ASUSWRT_GET_PATH):
+    async def __post(self, command, path=ASUSWRT_GET_PATH, *, retry=True):
         """
-        Private post method to execute a hook on the router and return the result
+        Private POST method to execute a hook on the router and return the result
 
         :param command: Command to send to the return
         :returns: string result from the router
         """
-        await self.async_connect()
         payload = f"hook={command}"
         try:
-            async with self._session.post(
-                url=self.__url(path),
-                headers=self._auth_headers,
-                data=payload,
-                timeout=self._timeout,
-                raise_for_status=True,
-                ssl=False,
-            ) as resp:
-                result = await resp.text()
-        except (asyncio.TimeoutError, OSError) as exc:
-            raise AsusWrtConnectionError(exc) from exc
-        except ConnectionRefusedError as exc:
-            self._auth_headers = None
-            raise AsusWrtConnectionError(exc) from exc
+            await self.async_connect()
+            result = await self.__http_post(
+                self.__url(path), self._auth_headers, payload
+            )
 
-        if result.find(ASUSWRT_ERROR_KEY, 0, len(ASUSWRT_ERROR_KEY)+5) >= 0:
+        except (AsusWrtConnectionError, AsusWrtClientResponseError):
+            if retry:
+                return await self.__post(command, path, retry=False)
+            raise
+
+        if result.find(ASUSWRT_ERROR_KEY, 0, len(ASUSWRT_ERROR_KEY) + 5) >= 0:
             self._auth_headers = None
+            if retry:
+                return await self.__post(command, path, retry=False)
             raise AsusWrtConnectionError("Not connected to the router")
 
         return result
@@ -216,6 +276,66 @@ class AsusWrtHttp:
         result = await self.__post(str(request), ASUSWRT_CMD_PATH)
         return result
 
+    @property
+    def hostname(self) -> str:
+        """Return the device hostname."""
+        return self._hostname
+
+    @property
+    def mac(self) -> str | None:
+        """Return the device mac address."""
+        return self._mac
+
+    @property
+    def is_connected(self) -> bool:
+        """Return if connection is active."""
+        return self._auth_headers is not None
+
+    async def async_disconnect(self):
+        """Close the managed session on exit."""
+        if self._managed_session and self._session is not None:
+            await self._session.close()
+            self._session = None
+        self._auth_headers = None
+
+    async def async_connect(self):
+        """Authenticate with the router."""
+        if self.is_connected:
+            return
+
+        if self._managed_session and self._session is None:
+            self._session = aiohttp.ClientSession()
+
+        auth = f"{self._username}:{self._password}".encode("ascii")
+        login_token = base64.b64encode(auth).decode("ascii")
+        payload = f"login_authorization={login_token}"
+        headers = {"user-agent": ASUSWRT_USR_AGENT}
+
+        result = await self.__http_post(
+            self.__url(ASUSWRT_LOGIN_PATH), headers, payload, get_json=True
+        )
+        if ASUSWRT_TOKEN_KEY not in result:
+            raise AsusWrtLoginError("Login Failed")
+
+        token = result[ASUSWRT_TOKEN_KEY]
+        self._auth_headers = {
+            "user-agent": ASUSWRT_USR_AGENT,
+            "cookie": f"{ASUSWRT_TOKEN_KEY}={token}",
+        }
+
+        # try to get the main properties after connect
+        await self._load_props()
+
+    async def _load_props(self) -> str | None:
+        """Load device properties from NVRam."""
+        if self._mac is not None:
+            return
+        try:
+            result = await self.async_get_settings(PROP_MAC_ADDR)
+        except AsusWrtError:
+            return
+        self._mac = result.get(PROP_MAC_ADDR)
+
     async def async_get_uptime(self):
         """
         Return uptime of the router
@@ -225,8 +345,8 @@ class AsusWrtHttp:
         :returns: JSON with last boot time and uptime in seconds
         """
         r = await self.__post(f"{CMD_UPTIME}()")
-        since = r.partition(':')[2].partition('(')[0]
-        up = int(r.partition('(')[2].partition(' ')[0])
+        since = r.partition(":")[2].partition("(")[0]
+        up = int(r.partition("(")[2].partition(" ")[0])
         return {"since": since, "uptime": up}
 
     async def async_get_memory_usage(self):
@@ -237,8 +357,8 @@ class AsusWrtHttp:
 
         :returns: JSON with memory variables
         """
-        s = json.loads(await self.__post(f"{CMD_MEMORY_USAGE}({PARAM_APPOBJ})"))
-        result = s[CMD_MEMORY_USAGE]
+        s = await self.__post(f"{CMD_MEMORY_USAGE}({PARAM_APPOBJ})")
+        result = _get_json_result(s, CMD_MEMORY_USAGE)
         return {k: int(v) for k, v in result.items()}
 
     async def async_get_cpu_usage(self):
@@ -250,8 +370,8 @@ class AsusWrtHttp:
 
         :returns: JSON with CPUs load statistics
         """
-        s = json.loads(await self.__post(f"{CMD_CPU_USAGE}({PARAM_APPOBJ})"))
-        result = s[CMD_CPU_USAGE]
+        s = await self.__post(f"{CMD_CPU_USAGE}({PARAM_APPOBJ})")
+        result = _get_json_result(s, CMD_CPU_USAGE)
         return {k: int(v) for k, v in result.items()}
 
     async def async_get_wan_info(self):
@@ -268,11 +388,11 @@ class AsusWrtHttp:
         """
         r = await self.__post(f"{CMD_WAN_INFO}()")
         status = {}
-        for f in r.split('\n'):
-            if 'return' in f:
+        for f in r.split("\n"):
+            if "return" in f:
                 if f"{CMD_WAN_INFO}_" in f:
-                    key = f.partition('(')[0].partition('_')[2]
-                    value = (f.rpartition(' ')[-1][:-2]).replace("'", "")
+                    key = f.partition("(")[0].partition("_")[2]
+                    value = (f.rpartition(" ")[-1][:-2]).replace("'", "")
                     status[key] = value
         return status
 
@@ -293,8 +413,8 @@ class AsusWrtHttp:
 
         :returns: JSON with a list of DHCP leases
         """
-        r = json.loads(await self.__post(f"{CMD_DHCP_LEASE}()"))
-        return r[CMD_DHCP_LEASE]
+        s = await self.__post(f"{CMD_DHCP_LEASE}()")
+        return _get_json_result(s, CMD_DHCP_LEASE)
 
     async def async_get_traffic_bytes(self):
         """
@@ -304,10 +424,20 @@ class AsusWrtHttp:
 
         :returns: JSON with sent and received bytes since last boot
         """
-        r = json.loads(await self.__post(f"{CMD_NET_TRAFFIC}({PARAM_APPOBJ})"))
-        meas = r[CMD_NET_TRAFFIC]
-        rx = int(meas["INTERNET_rx"], base=16)
-        tx = int(meas["INTERNET_tx"], base=16)
+        s = await self.__post(f"{CMD_NET_TRAFFIC}({PARAM_APPOBJ})")
+        meas = _get_json_result(s, CMD_NET_TRAFFIC)
+        traffic = None
+        if "INTERNET_rx" in meas:
+            traffic = "INTERNET"
+        elif "BRIDGE_rx" in meas:
+            traffic = "BRIDGE"
+
+        if traffic:
+            rx = int(meas[f"{traffic}_rx"], base=16)
+            tx = int(meas[f"{traffic}_tx"], base=16)
+        else:
+            rx = tx = 0
+
         return {"rx": rx, "tx": tx}
 
     async def async_get_traffic_rates(self):
@@ -368,7 +498,7 @@ class AsusWrtHttp:
         for s in setting_list:
             resp = await self.__post(_nvram_cmd(s))
             if resp:
-                result[s] = json.loads(resp)[s]
+                result[s] = _get_json_result(resp, s)
         return result
 
     async def async_get_clients_fullinfo(self) -> list[dict[str, any]]:
@@ -394,7 +524,8 @@ class AsusWrtHttp:
                 ]
         :returns: JSON with list of mac address and all client related info
         """
-        result = json.loads(await self.__post(f"{CMD_CLIENT_LIST}()"))
+        s = await self.__post(f"{CMD_CLIENT_LIST}()")
+        result = _get_json_result(s)
         return [result.get(CMD_CLIENT_LIST, {})]
 
     async def async_get_connected_mac(self):
@@ -408,7 +539,7 @@ class AsusWrtHttp:
         lst = [
             mac
             for mac, info in clnts[0].items()
-            if len(mac) == 17 and info.get("isOnline", '0') == '1'
+            if len(mac) == 17 and info.get("isOnline", "0") == "1"
         ]
         return lst
 
@@ -420,25 +551,38 @@ class AsusWrtHttp:
         :return: JSON dict with mac as key and a namedtuple with mac, ip address and name as value
         """
         clnts = await self.async_get_clients_fullinfo()
-        result = {}
+        dev_list = []
+        mesh_nodes = {self._mac: self._hostname} if self._mac else {}
         for mac, info in clnts[0].items():
-            if len(mac) == 17 and info.get("isOnline", '0') == '1':
+            if len(mac) == 17 and info.get("isOnline", "0") == "1":
+                if is_mesh_node := info.get("amesh_isRe", "0") == "1":
+                    mesh_nodes[mac] = info.get("ip")
                 if not (name := info.get("nickName")):
                     name = info.get("name")
-                result[mac] = Device(mac, info.get("ip"), name)
-                # lst.append(
-                #     {
-                #         "name": clnts['get_clientlist'][c]['name'],
-                #         "nickName": clnts['get_clientlist'][c]['nickName'],
-                #         "ip": clnts['get_clientlist'][c]['ip'],
-                #         "mac": clnts['get_clientlist'][c]['mac'],
-                #         "isOnline": clnts['get_clientlist'][c]['isOnline'],
-                #         "curTx": clnts['get_clientlist'][c]['curTx'],
-                #         "curRx": clnts['get_clientlist'][c]['curRx'],
-                #         "totalTx": clnts['get_clientlist'][c]['totalTx'],
-                #         "totalRx": clnts['get_clientlist'][c]['totalRx'],
-                #     }
-                # )
+                is_wl = info.get("isWL", "0") != "0"
+                dev_list.append(
+                    Device(
+                        mac,
+                        info.get("ip"),
+                        name,
+                        info.get("amesh_papMac"),
+                        is_mesh_node,
+                        is_wl,
+                    )
+                )
+
+        result = {}
+        for dev in dev_list:
+            node_ip = mesh_nodes.get(dev.node) if dev.node else None
+            result[dev.mac] = Device(
+                dev.mac,
+                dev.ip,
+                dev.name,
+                node_ip or self._hostname,
+                dev.is_mesh_node,
+                dev.is_wl,
+            )
+
         return result
 
     async def async_get_client_info(self, client_mac):
